local MAJOR = "LibScriptableDisplayWidgetGestures-1.0" 
local MINOR = 14

assert(LibStub, MAJOR.." requires LibStub") 
local WidgetGestures = LibStub:NewLibrary(MAJOR, MINOR)
if not WidgetGestures then return end
local LibWidget = LibStub("LibScriptableDisplayWidget-1.0", true)
assert(LibWidget, MAJOR .. " requires LibScriptableDisplayWidget-1.0")
local LibTimer = LibStub("LibScriptableDisplayTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires LibScriptableDisplayTimer-1.0")
local Evaluator = LibStub("LibScriptableDisplayEvaluator-1.0", true)
assert(Evaluator, MAJOR .. " requires LibScriptableDisplayEvaluator-1.0")
local PluginUtils = LibStub("LibScriptableDisplayPluginUtils-1.0", true)
assert(PluginUtils, MAJOR .. " requires LibScriptableDisplayPluginUtils-1.0")
local LibMouseGestures = LibStub("LibMouseGestures-1.0", true)
assert(LibMouseGestures, MAJOR .. " requires LibMouseGestures-1.0")

if not WidgetGestures.__index then WidgetGestures.__index = WidgetGestures end

WidgetGestures.defaults = {
	repeating = true,
	update = 1000,
}

local widgetType = {gesture=true}

local pool = setmetatable({}, {__mode = "v"})
local function new(...)
	local obj = table.remove(pool) or {}
	for i = 1, select("#", ...) do
		obj[i] = select(i, ...)
	end
	return obj
end

local function newDict(...)
	local obj = table.remove(pool) or {}
	for i = 1, select("#", ...), 2 do
		local key = select(i, ...)
		local val = select(i + 1, ...)
		obj[key] = val
	end
	return obj
end

local function del(obj)
	wipe(obj)
	tinsert(pool, obj)
end

local cache = setmetatable({}, {__mode = "v"})
local function newRec()
	local rec = LibMouseGestures:New(UIParent);
	return rec
end

local function delRec(rec)
	tinsert(cache, rec)
end

local pool = setmetatable({}, {__mode = "k"})

-- Create a new LibScriptableDisplayWidgetGestures object
-- @name WidgetGestures.New
-- @usage WidgetGestures:New(visitor, name, config, errorLevel)
-- @param visitor An LibScriptableDisplayCore-1.0 object, or provide your own
-- @param name A name for the timer widget
-- @param config This timeer's parameters
-- @param errorLevel The errorLevel for this object
-- @param callback An optional callback function to be executed when the gesture is performed.
-- @param An optional timer. This should have a :Start() and :Stop().
-- @return A new LibScriptableDisplayWidgetGestures widget
function WidgetGestures:New(visitor, name, config, errorLevel, callback, timer) 
	assert(name, "WidgetGestures requires a name.")
	assert(config, "Please provide the timer with a config")
	assert(config.expression, name .. ": Please provide the timer with an expression")
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
		obj.options = {}
	end
		
	setmetatable(obj, self)

	obj.widget = LibWidget:New(obj, visitor, MAJOR .. "." .. name, config, 0, 0, 0, widgetType, errorLevel)
	
	obj.expression = config.expression
	obj.repeating = config.repeating or WidgetGestures.defaults.repeating
	obj.update = config.update or WidgetGestures.defaults.update
	obj.disabled = config.disabled
	obj.gestures = config.gestures
	obj.callback = callback
	obj.error = LibError:New(MAJOR .. ": " .. name, errorLevel)
		
	obj.timer = timer or LibTimer:New("WidgetGestures.timer " .. obj.name, obj.update, obj.repeating, self.Update, obj, obj.errorLevel)
	
	return obj	
end


-- Delete a LibScriptableDisplayWidgetGestures object
-- @name LibScriptableDisplayWidgetGestures.Del
-- @usage LibScriptableDisplayWidgetGestures:Del([marq])
-- @param marq A LibScriptableDisplayWidgetGestures object
-- @return Nothing
function WidgetGestures:Del(timer)
	if not timer then
		timer = self
	end
	timer:Stop()
	timer.widget:Del()
	timer.timer:Del()
	pool[timer] = true
end

-- Start a LibScriptableDisplayWidgetGestures
-- @name LibScriptableDisplayWidgetGestures.Start
-- @usage object:Start()
-- @return Nothing
function WidgetGestures:Start()
	if self.update > 0 and not self.disabled and not self.active then
		self.timer:Start()
		self.active = true
	end
end

-- Stop a LibScriptableDisplayWidgetGestures
-- @name LibScriptableDisplayWidgetGestures.Stop
-- @usage object:Stop()
-- @return Nothing
function WidgetGestures:Stop()
	self.timer:Stop()
	self.active = false
end

local function stopFunc(rec)
	local g = rec:GetGist();
	local self = rec.widgetdata
	
	local comparisons = new()
		
	if ( #g ) then
		for n = 1, #g do
			if n > #self.gestures then break end
			local type = g[n][1]
			local pattern = g[n][2]
			tinsert(comparisons, newDict("type", new(self.gestures[n].type, type), "pattern", new(self.gestures[n].pattern, pattern)))
		end
	end
	if #comparisons > 0 then
		local execute = false
		local current = comparisons[1]
		for i, v in ipairs(comparisons) do 
			if v.type[1] == v.type[2] and v.pattern[1] == v.pattern[2] then
				execute = true
				current = v
			elseif current.type[1] == v.type[2] and current.pattern[1] == v.pattern[2] then
				execute = true
				current = v
			else
				execute = false
			end
		end
		if #comparisons < #self.gestures then execute = false end
		if execute then
			self.visitor.environment.self = self
			Evaluator.ExecuteCode(self.visitor.environment, self.name, self.expression)
			self.visitor.environment.self = nil		
			if type(self.callback) == "function" then
				self:callback()
			end
			self.error:Print("Execute", 2)
		end
	end
	for i, v in ipairs(comparisons) do
		for k, v in pairs(v) do
			del(v)
		end
		del(v)
	end
	del(comparisons)
		
	--[[rec:StartCapture({
		startButton		= "Freehand",
		stopButton		= "LeftButtonUp",
		cancelButton    = "RightButtonDown",
		showTrail 		= false,
		stopFunc		= stopFunc
	})]]
	
	self.waiting = false

	delRec(rec)
end

local capture = {
	startButton		= "Freehand",
	stopButton		= "LeftButtonUp",
	cancelButton    = "RightButtonDown",
	showTrail 		= false,
	stopFunc		= stopFunc
}

-- Update data
-- @name LibScriptableDisplayWidgetGestures.Update
-- @name LibScriptableDisplayWidgetGestures.Update
-- @usage LibScriptableDisplayWidgetGestures:Update()
-- @return Nothing
function WidgetGestures:Update()
	if self.waiting then return end

	local rec = newRec()
	rec.widgetdata = self
	
	rec:StartCapture(capture)
	
	self.waiting = true
end

function WidgetGestures:GetOptions(visitor, db)
		local defaults = WidgetGestures.defaults
		local options = {
			enabled = {
				name = "Enabled",
				desc = "Whether this timer is enabled or not",
				type = "toggle",
				get = function() return db.enabled end,
				set = function(info, v) 
					db.enabled = v
					if visitor.RestartTimers then
						visitor:RestartTimers()
					end					
					db["enabledDirty"] = true
				end,
				order = 2
			},
			update = {
				name = "Update Rate",
				desc = "Enter the timer's refresh rate",
				type = "input",
				pattern = "%d",
				get = function()
					return tostring(db.update or defaults.update)
				end,
				set = function(info, v)
					db.update = tonumber(v)
					if visitor.RestartTimers then
						visitor:RestartTimers()
					end
					db["updateDirty"] = true
				end,
				order = 3
			},
			expression = {
				name = "Expression",
				desc = "Enter this widget's expression",
				type = "input",
				width = "full",
				multiline = true,
				get = function()
					return db.expression or defaults.expression
				end,
				set = function(info, v)
					db.expression = v
					if visitor.RestartTimers then
						visitor:RestartTimers()
					end
					db["expressionDirty"] = true
				end,
				order = 5
			},
			
		}
	return options
end
